﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Microsoft.StylusInput;
using MidiStylusLibrary;
using System.Threading;
using NAudio.Midi;

namespace MidiStylus
{
    public partial class MainForm : Form
    {
        #region Variables

        private RealTimeStylus _stylus;
        private StylusData _data;
        private MidiDevice _midiDevice;
        private Configuration _configuration = new Configuration();
        private bool _isRunning = false;
        private bool _isReady = true;
        private int _lastX = 0;
        private int _lastY = 0;
        private int _lastPressure = 0;

        #endregion

        #region Properties

        public MidiDevice MidiInterface
        {
            get { return _midiDevice; }
        }

        public bool IsReady
        {
            get { return _isReady; }
        }

        #endregion

        #region Constructors

        public MainForm()
        {
            InitializeComponent();

            this.Initialize();
        }

        #endregion

        #region Methods

        public void Initialize()
        {
            // Load configuration data from XML
            try
            {
                _configuration = new Configuration("Settings.xml");
            }
            catch (Exception ex)
            {
                MessageBox.Show("No configuration file found, using default settings", "No configuration file", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }

            InitializeStylus();
            InitializeMIDI();

            // Set up the form controls
            UpdateStatusBar(0, 0, 0);

            bwUpdater.WorkerReportsProgress = true;
            bwUpdater.WorkerSupportsCancellation = true;
        }

        private void StartRunning()
        {
            if (this._isReady)
            {
                _isRunning = true;
                runToolStripMenuItem.Text = "Stop";

                // Start the running thread
                bwUpdater.RunWorkerAsync();
            }
            else
            {
                MessageBox.Show("MidiStylus is not configured.  Click on Settings and set up your hardware to continue.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void StopRunning()
        {
            _isRunning = false;
            runToolStripMenuItem.Text = "Run";

            // Stop the processing thread
            bwUpdater.CancelAsync();
        }

        private void UpdateStatusBar(int x, int y, int pressure)
        {
            // Update the screen position indicators
            lblPressure.Text = "Pressure (" + _configuration.PressureControlNumber.ToString() + "): " + pressure.ToString();
            lblXValue.Text = "X (" + _configuration.XControlNumber.ToString() + "): " + x.ToString();
            lblYValue.Text = "Y (" + _configuration.YControlNumber.ToString() + "): " + y.ToString();
        }

        public void InitializeStylus()
        {
            try
            {
                // We create the stylus
                this._stylus = new RealTimeStylus(pbWorkspace);
                this._data = new StylusData(pbWorkspace);

                this._stylus.SyncPluginCollection.Add(_data);
                this._stylus.Enabled = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message.ToString(), "Error initializing stylus", MessageBoxButtons.OK, MessageBoxIcon.Error);
                this._isReady = false;
            }
        }

        public void InitializeMIDI()
        {
            try
            {
                if (this._midiDevice != null)
                {
                    this._midiDevice.CloseMIDI();
                }

                _midiDevice = new MidiDevice(_configuration.DeviceNumber);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message.ToString(), "Error initializing MIDI device", MessageBoxButtons.OK, MessageBoxIcon.Error);
                this._isReady = false;
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);
        }

        #endregion

        #region Event Handlers

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                AboutForm about = new AboutForm();
                about.ShowDialog();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message.ToString(), "Program Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void settingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                this.StopRunning();

                SettingsForm settings = new SettingsForm();
                settings.Owner = this;

                if (settings.ShowDialog() == DialogResult.OK)
                {
                    this.Initialize();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message.ToString(), "Program Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void bwUpdater_DoWork(object sender, DoWorkEventArgs e)
        {
            // Runs as long as the program is running
            while (this._isRunning)
            {
                // Force the progress changed event
                bwUpdater.ReportProgress(0);
                Thread.Sleep(50);
            }
        }

        private void bwUpdater_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            int x = 0;
            int y = 0;
            int pressure = 0;

            if (this._data != null)
            {
                x = _data.XValue;
                y = _data.YValue;
                pressure = _data.Pressure;
            }

            // Normalize values
            if (x >= pbWorkspace.Width) x = pbWorkspace.Width;
            if (y >= pbWorkspace.Height) y = pbWorkspace.Height;
            if (x <= 0) x = 0;
            if (y <= 0) y = 0;

            // Map and transmit MIDI
            int channel = _configuration.ChannelNumber;
            x = StylusData.NormalizeValue(x, pbWorkspace.Width, 127);
            y = StylusData.NormalizeValue(y, pbWorkspace.Height, 127);
            pressure = StylusData.NormalizeValue(pressure, 32767, 127);

            // Now we check to see if these are new values, and if so, we send the MIDI message
            if (x == _lastX && y == _lastY && pressure == _lastPressure)
            {
                // Don't do anything if the values haven't changed
            }
            else
            {
                _midiDevice.ControlChange(channel, _configuration.XControlNumber, x);
                _midiDevice.ControlChange(channel, _configuration.YControlNumber, y);
                _midiDevice.ControlChange(channel, _configuration.PressureControlNumber, pressure);

                // Update the screen position indicators
                UpdateStatusBar(x, y, pressure);
            }

            _lastX = x;
            _lastY = y;
            _lastPressure = pressure;
        }

        private void runToolStripMenuItem_Click_1(object sender, EventArgs e)
        {
            if (this._isRunning)
            {
                StopRunning();
            }
            else
            {
                StartRunning();
            }
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this._midiDevice != null)
            {
                this._midiDevice.CloseMIDI();
            }

            if (this._isRunning)
            {
                StopRunning();
            }
        }

        #endregion
    }
}
